# Sphinx extensions for generating HTML output

extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.extlinks',
              'sphinx.ext.intersphinx',
              'sphinx.ext.mathjax',
              'sphinx.ext.napoleon',
              'sphinx.ext.todo',
              'IPython.sphinxext.ipython_console_highlighting',
              'jupyter_sphinx',
              'sphinx_gallery.gen_gallery']
